/**********************************************************************
**
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Designer.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

#include "xplodesyntaxhighlighter.h"

// This file has been slightly modified for Xplode

#include "paragdata.h"
#include "xplodeshared.h"
#include "plasmkeywords.h"

extern PlasmKeywords *plasmKeywords;

// Keyword/Format mapping
static QMap<int, QMap<QString, int> > *wordMap = 0;

//  Constructor
XplodeSyntaxHighlighter::XplodeSyntaxHighlighter() : QTextPreProcessor(), lastFormat(0), lastFormatId(-1)
{
        // Add a format for each style
        addFormat(Standard, new QTextFormat(*plasmEditorFont, Qt::black));
        addFormat(Comment, new QTextFormat(*plasmEditorFont, Qt::gray));
        addFormat(Number, new QTextFormat(*plasmEditorFont, Qt::red));
        addFormat(String, new QTextFormat(*plasmEditorFont, Qt::darkCyan));
        addFormat(Type, new QTextFormat(*plasmEditorFont, Qt::blue));
        addFormat(Keyword, new QTextFormat(*plasmEditorFont, Qt::blue));
        // Initialize mapping
        if (wordMap) return;
        wordMap = new QMap<int, QMap<QString, int> >;
        int len, i;
        i = 0;
        // Cycle in all definitions and update the tree
        QMap<QString, KeyWord*>::Iterator it;
        for (it = plasmKeywords->db.begin(); it != plasmKeywords->db.end(); ++it)
        {
                len = it.data()->getName().length();
                if (!wordMap->contains(len)) wordMap->insert(len, QMap<QString, int>());
                QMap<QString, int> &map = wordMap->operator[](len);
                map[it.data()->getName()] = Keyword;
                i++;
        }
}

// Destructor
XplodeSyntaxHighlighter::~XplodeSyntaxHighlighter()
{
}

// Map a string into its corresponding format
static int string2Id(const QString &s)
{
        if (s == "Standard") return XplodeSyntaxHighlighter::Standard;
        if (s == "Comment") return XplodeSyntaxHighlighter::Comment;
        if (s == "Number") return XplodeSyntaxHighlighter::Number;
        if (s == "String") return XplodeSyntaxHighlighter::String;
        if (s == "Type") return XplodeSyntaxHighlighter::Type;
        if (s == "Preprocessor") return XplodeSyntaxHighlighter::PreProcessor;
        if (s == "Keyword") return XplodeSyntaxHighlighter::Keyword;
        return XplodeSyntaxHighlighter::Standard;
}

// Update styles
void XplodeSyntaxHighlighter::updateStyles(const QMap<QString, ConfigStyle> &styles)
{
        for (QMap<QString, ConfigStyle>::ConstIterator it = styles.begin(); it != styles.end(); ++it)
        {
                int id = string2Id(it.key());
                QTextFormat *f = format(id);
                if (!f) continue;
                f->setFont((*it).font);
                f->setColor((*it).color);
        }
}

// Process the text and formats it.
// The process is a finite-state machine computation
void XplodeSyntaxHighlighter::process(QTextDocument *doc, QTextParagraph *string, int, bool invalidate)
{
        // Initialize formats
        QTextFormat *formatStandard = format(Standard);
        QTextFormat *formatComment = format(Comment);
        QTextFormat *formatNumber = format(Number);
        QTextFormat *formatString = format(String);
        QTextFormat *formatType = format(Type);
        QTextFormat *formatPreProcessor = format(PreProcessor);

        // States
        const int StateStandard = 0;
        const int StateCommentStart1 = 1;
        const int StateCCommentStart2 = 2;
        const int StateCppCommentStart2 = 3;
        const int StateCComment = 4;
        const int StateCppComment = 5;
        const int StateCCommentEnd1 = 6;
        const int StateCCommentEnd2 = 7;
        const int StateStringStart = 8;
        const int StateString = 9;
        const int StateStringEnd = 10;
        const int StateString2Start = 11;
        const int StateString2 = 12;
        const int StateString2End = 13;
        const int StateNumber = 14;
        const int StatePreProcessor = 15;

        // Tokens
        const int InputAlpha = 0;
        const int InputNumber = 1;
        const int InputParen = 4;
        const int InputSpace = 5;
        const int InputPercent = 7;
        const int InputApostrophe = 8;
        const int InputSep = 9;

        // State/Input Matrix: states (rows), inputs (columns)
        static uchar table[16][10] =
        {
                { StateStandard,      StateNumber,     StateStandard,       StateCommentStart1,    StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard }, // StateStandard
                { StateStandard,      StateNumber,   StateCCommentStart2, StateCppCommentStart2, StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard }, // StateCommentStart1
                { StateCComment,      StateCComment,   StateCCommentEnd1,   StateCComment,         StateCComment,   StateCComment,   StateCComment,     StateCComment,    StateCComment,     StateCComment }, // StateCCommentStart2
                { StateCppComment,    StateCppComment, StateCppComment,     StateCppComment,       StateCppComment, StateCppComment, StateCppComment,   StateCppComment,  StateCppComment,   StateCppComment }, // CppCommentStart2
                { StateCComment,      StateCComment,   StateCCommentEnd1,   StateCComment,         StateCComment,   StateCComment,   StateCComment,     StateCComment,    StateCComment,     StateCComment }, // StateCComment
                { StateCppComment,    StateCppComment, StateCppComment,     StateCppComment,       StateCppComment, StateCppComment, StateCppComment,   StateCppComment,  StateCppComment,   StateCppComment }, // StateCppComment
                { StateCComment,      StateCComment,   StateCCommentEnd1,   StateCCommentEnd2,     StateCComment,   StateCComment,   StateCComment,     StateCComment,    StateCComment,     StateCComment }, // StateCCommentEnd1
                { StateStandard,      StateNumber,     StateStandard,       StateCommentStart1,    StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard }, // StateCCommentEnd2
                { StateString,        StateString,     StateString,         StateString,           StateString,     StateString,     StateString,       StateStringEnd,   StateString,       StateString }, // StateStringStart
                { StateString,        StateString,     StateString,         StateString,           StateString,     StateString,     StateString,       StateStringEnd,   StateString,       StateString }, // StateString
                { StateStandard,      StateStandard,   StateStandard,       StateCommentStart1,    StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard }, // StateStringEnd
                { StateString2,       StateString2,    StateString2,        StateString2,          StateString2,    StateString2,    StateString2,      StateString2,     StateString2End,   StateString2 }, // StateString2Start
                { StateString2,       StateString2,    StateString2,        StateString2,          StateString2,    StateString2,    StateString2,      StateString2,     StateString2End,   StateString2 }, // StateString2
                { StateStandard,      StateStandard,   StateStandard,       StateCommentStart1,    StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard }, // StateString2End
                { StateNumber,        StateNumber,     StateStandard,       StateCommentStart1,    StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard }, // StateNumber
                { StatePreProcessor,  StateStandard,   StateStandard,       StateCommentStart1,    StateStandard,   StateStandard,   StatePreProcessor, StateStringStart, StateString2Start, StateStandard } // StatePreProcessor
    };

        QString buffer;

        // Initialization
        int state = StateStandard;
        if (string->prev())
        {
                if (string->prev()->endState() == -1) process(doc, string->prev(), 0, false);
                state = string->prev()->endState();
        }
        int input;
        int i = 0;
        bool lastWasBackSlash = false;
        bool makeLastStandard = false;

        ParagData *paragData = (ParagData*)string->extraData();
        if (paragData)
                paragData->parenList.clear();
        else
                paragData = new ParagData;

        // Tokens
        static QString alphabeth = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
        static QString mathChars = "xXeE";
        static QString numbers = "0123456789";
        bool questionMark = false;
        QChar lastChar;
        // Parsing loop
        for (;;)
        {
                QChar c = string->at(i)->c.lower();
                // Check the input character
                switch (c)
                {
                        // An open parenthesis
                        case '(': case '[': case '<':
                                input = InputParen;
                                if (state == StateStandard ||
                                        state == StateNumber ||
                                        state == StatePreProcessor ||
                                        state == StateCCommentEnd2 ||
                                        state == StateCCommentEnd1 ||
                                        state == StateString2End ||
                                        state == StateStringEnd) paragData->parenList << Paren(Paren::Open, c, i);
                                break;
                        // A closing parenthesis
                        case ')': case ']': case '>':
                                input = InputParen;
                                if (state == StateStandard ||
                                        state == StateNumber ||
                                        state == StatePreProcessor ||
                                        state == StateCCommentEnd2 ||
                                        state == StateCCommentEnd1 ||
                                        state == StateString2End ||
                                        state == StateStringEnd) paragData->parenList << Paren(Paren::Closed, c, i);
                                break;
                        // A comment token
                        case '%':
                                input = InputPercent;
                                break;
                        // A string token
                        case '\'':
                                input = InputApostrophe;
                                break;
                        // A white space
                        case ' ':
                                input = InputSpace;
                                break;
                        // A number
                        case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9': case '0':
                                if (alphabeth.find(lastChar) != -1 && (mathChars.find(lastChar) == -1 || numbers.find(string->at(i - 1)->c) == -1))
                                        input = InputAlpha;
                                else
                                {
                                        if (input == InputAlpha && numbers.find(lastChar) != -1)
                                                input = InputAlpha;
                                        else
                                                input = InputNumber;
                                }
                                break;
                        // Any other char
                        default:
                        {
                                if (!questionMark && c == '?') questionMark = true;
                                if (c.isLetter() || c == '_')
                                        input = InputAlpha;
                                else
                                        input = InputSep;
                        } break;
                }
                lastWasBackSlash = !lastWasBackSlash && c == '\\';
                if (input == InputAlpha) buffer += c;
                // Map the input character and current state into the new state
                state = table[state][input];
                // Format the text depending on the current state
                switch (state)
                {
                        // Standard text
                        case StateStandard:
                        {
                                int len = buffer.length();
                                string->setFormat(i, 1, formatStandard, false);
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                if (buffer.length() > 0 && input != InputAlpha)
                                {
                                        if ( buffer[0] == 'Q')
                                        {
                                                string->setFormat(i - buffer.length(), buffer.length(), formatType, false);
                                        }
                                        else
                                        {
                                                QMap<int, QMap<QString, int> >::Iterator it = wordMap->find(len);
                                                if (it != wordMap->end())
                                                {
                                                        QMap<QString, int>::Iterator it2 = (*it).find(buffer);
                                                        if (it2 != (*it).end())
                                                                string->setFormat(i - buffer.length(), buffer.length(), format((*it2)), false);
                                                }
                                        }
                                        buffer = QString::null;
                                }
                        } break;
                        // Comment: unused
                        case StateCommentStart1:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = true;
                                buffer = QString::null;
                                break;
                        // C Comment: unused
                        case StateCCommentStart2:
                                string->setFormat(i - 1, 2, formatComment, false);
                                makeLastStandard = false;
                                buffer = QString::null;
                                break;
                        // C++ Comment: unused
                        case StateCppCommentStart2:
                                string->setFormat(i - 1, 2, formatComment, false);
                                makeLastStandard = false;
                                buffer = QString::null;
                                break;
                        // C Comment: unused
                        case StateCComment:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatComment, false);
                                buffer = QString::null;
                                break;
                        // C++ Comment
                        case StateCppComment:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatComment, false);
                                buffer = QString::null;
                                break;
                        // C Comment: unused
                        case StateCCommentEnd1:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatComment, false);
                                buffer = QString::null;
                                break;
                        // C Comment: unused
                        case StateCCommentEnd2:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatComment, false);
                                buffer = QString::null;
                                break;
                        // String start
                        case StateStringStart:
                                if (makeLastStandard ) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatStandard, false);
                                buffer = QString::null;
                                break;
                        // String
                        case StateString:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                // WAS UNCOMMENTED: string->setFormat( i, 1, formatString, false );
                                string->setFormat( i, 1, formatComment, false );
                                buffer = QString::null;
                                break;
                        // String end
                        case StateStringEnd:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatStandard, false);
                                buffer = QString::null;
                                break;
                        // String start
                        case StateString2Start:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatStandard, false);
                                buffer = QString::null;
                                break;
                        // String
                        case StateString2:
                                if (makeLastStandard ) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatString, false);
                                buffer = QString::null;
                                break;
                        // String end
                        case StateString2End:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatStandard, false);
                                buffer = QString::null;
                                break;
                        // Number
                        case StateNumber:
                                if (makeLastStandard) string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatNumber, false);
                                buffer = QString::null;
                                break;
                        // C/C++ Preprocessor: unused
                        case StatePreProcessor:
                                if (makeLastStandard)
                                string->setFormat(i - 1, 1, formatStandard, false);
                                makeLastStandard = false;
                                string->setFormat(i, 1, formatPreProcessor, false);
                                buffer = QString::null;
                                break;
                }
                lastChar = c;
                i++;
                if (i >= string->length()) break;
    }
        string->setExtraData(paragData);
        if (state == StateCComment ||
                state == StateCCommentEnd1)
        {
                string->setEndState(StateCComment);
    }
        else
                if (state == StateString)
                {
                        string->setEndState(StateString);
                }
                else
                        if (state == StateString2)
                        {
                                string->setEndState(StateString2);
                        }
                        else
                        {
                                string->setEndState(StateStandard);
                        }

        string->setFirstPreProcess(false);
        if (invalidate && string->next() && !string->next()->firstPreProcess() && string->next()->endState() != -1)
        {
                QTextParagraph *p = string->next();
                while (p)
                {
                        if (p->endState() == -1) return;
                        p->setEndState(-1);
                        p = p->next();
                }
        }
}

// Return the text format
QTextFormat *XplodeSyntaxHighlighter::format(int id)
{
        if (lastFormatId == id  && lastFormat) return lastFormat;
        QTextFormat *f = formats[id];
        lastFormat = f ? f : formats[0];
        lastFormatId = id;
        return lastFormat;
}

// Add a text format
void XplodeSyntaxHighlighter::addFormat(int id, QTextFormat *f)
{
        formats.insert(id, f);
}

// Remove a text format
void XplodeSyntaxHighlighter::removeFormat(int id)
{
        formats.remove(id);
}

